#!/usr/bin/bash

CURR_TIME=$(date '+%d-%m-%y %H-%M-%S')
export CURR_TIME
export LOG_FILE='/var/log/join_to_domain.log'
export SCRIPT_VERSION='1.0.0'

SAVE_SUFFIX='jtd_save'
HOSTS_FILE='/etc/hosts'
HOSTNAME_FILE='/etc/hostname'
SSH_CONFIG_FILE='/etc/ssh/ssh_config'
SSHD_CONFIG_FILE='/etc/ssh/sshd_config'
CHRONY_CONF_FILE='/etc/chrony.conf'
KRB5_CONF_FILE='/etc/krb5.conf'
CRYPTO_POLICIES_FILE='/etc/krb5.conf.d/crypto-policies'
SMB_CONF_FILE='/etc/samba/smb.conf'
WINBIND_NSSWITCH_CONF_FILE='/etc/authselect/custom/winbind_domain/nsswitch.conf'
SSSD_NSSWITCH_CONF_FILE='/etc/authselect/custom/sssd_domain/nsswitch.conf'
USER_NSSWITCH_CONF_FILE='/etc/authselect/user-nsswitch.conf'
LIMITS_FILE='/etc/security/limits.conf'
SSSD_CONF_FILE='/etc/sssd/sssd.conf'
WINBIND_CONF_FILE='/etc/security/pam_winbind.conf'
SELINUX_CONFIG_FILE='/etc/selinux/config'

CH_FILES="$HOSTS_FILE $HOSTNAME_FILE $SSH_CONFIG_FILE $SSHD_CONFIG_FILE $CHRONY_CONF_FILE $KRB5_CONF_FILE $CRYPTO_POLICIES_FILE $SMB_CONF_FILE $WINBIND_NSSWITCH_CONF_FILE $SSSD_NSSWITCH_CONF_FILE $USER_NSSWITCH_CONF_FILE $LIMITS_FILE $SSSD_CONF_FILE $WINBIND_CONF_FILE $SELINUX_CONFIG_FILE"
CURR_HOSTNAME='' CURR_DOMAIN='' AUTO_YES=''

S_LANG=$LANG
export LANG=C.UTF-8

CURR_HOSTNAME=$(hostname -s)
CURR_DOMAIN=$(hostname -d)
[[ -z "$CURR_DOMAIN" ]] && CURR_DOMAIN=$(cat /etc/resolv.conf | awk '/^search/ && !/^#/{print $2}')
export CURR_HOSTNAME
export CURR_DOMAIN

. /usr/bin/gettext.sh
TEXTDOMAIN='join-to-domain'
export TEXTDOMAIN

# ------------------------------------------------------------------------------
GetText()
{
  LANG="$S_LANG"
  eval_gettext "$@"
  export LANG=C.UTF-8
}

# ------------------------------------------------------------------------------
Usage()
{
  local PROGNAME
  PROGNAME=$(basename $0)

  cat <<EOF

$(GetText "Allows you to add a computer to Windows domain (2008/2012/2016/2019/2022), SAMBA or IPA domain.")
$(GetText "Must be run with root user rights.")

  $(GetText "Usage"):

    $(GetText "Add a computer to Windows or IPA domain"):

EOF
  if [[ -z "$CURR_DOMAIN" ]]; then
    cat <<EOF
      $PROGNAME -d | --domain <$(GetText "domain name")>  -u | --admin <$(GetText "domain admin name")> [$(GetText "option")]
      $(GetText "option"):

EOF
  else
    cat <<EOF
    $PROGNAME  -u | --admin <$(GetText "domain admin name")> [$(GetText "option")]

    $(GetText "option"):

      -d | --domain <$(GetText "domain name")>  ($(GetText "default"): ${CURR_DOMAIN})
EOF
  fi
  cat <<EOF
      -n | --pc-name <$(GetText "computer name")> ($(GetText "default"): ${CURR_HOSTNAME})
      -o | --ou <ou>          - $(GetText "Organization unit, bottom-up order").
                                         $(GetText "Example: OU=MyPC,OU=ITdepartment").
      -c | --dc-name <$(GetText "domain controller name")>
      -w | --winbind          - $(GetText "Using Winbind (SSSD is used by default)" ).
      -W | --wg <$(GetText "domain name")>  - $(GetText "Domain name before Windows 2000").
      -y | --yes              - $(GetText "Answer yes to all questions").
      -f | --force            - $(GetText "Forced entry into domain under your previous PC name").
      -i | --lower-case       - $(GetText "Username in lowercase").
      -s | --save-case        - $(GetText "Username with save case, as in AD").
      -a | --dns-auth-none    - $(GetText "When dynamically registering a DNS address, authentication on the DNS server is not required").
      -k | --ssh-krb5         - $(GetText "Enable support for authentication of domain users using kerberos on the SSH server").
EOF

  cat <<EOF

    $(GetText "Delete PC from a domain"):
        $PROGNAME -D | --delete-pc  -u | --admin <$(GetText "domain admin name")> [  -d | --domain <$(GetText "domain name")> ] [ -n | --pc-name <$(GetText "computer name")> ]

    $(GetText "Delete sssd cache"):
        $PROGNAME -r | --del-sssd-cache

    $(GetText "Display this help and exit"):
        $PROGNAME -h | --help

    $(GetText "Display this version and exit"):
        $PROGNAME -v | --version

EOF
}

# ------------------------------------------------------------------------------
LogMess()
{
    local MESS='' PRINT='' PRINT_ONLY='' DELIM='' ERR='' WARN='' OPTIND=1 COLOR=''

    while getopts "wem:dDpPS" OPTION; do
        case $OPTION in
            'm') MESS="$OPTARG" ;;
            'p') PRINT="Y" ;;
            'e') ERR="Y" ;;
            'w') WARN="Y" ;;
            'P') PRINT_ONLY="Y" ;;
            'd') DELIM="-------------------------------------" ;;
            'D') DELIM="------------------------------------------------------------------------------" ;;
            'S') echo "################################################################################" >>"$LOG_FILE" ;;
            *) echo "LogMess() Invalid option: $*" ;;
        esac
    done
    shift $((OPTIND-1))
    OPTIND=1

    if [[ -n "$DELIM" ]]; then
       [[ -n "$PRINT_ONLY" ]] && echo "$DELIM" || echo "$DELIM" >>"$LOG_FILE"
       [[ -n "$PRINT" ]] && echo "$DELIM"
    fi

    [[ -z "$MESS" ]] && return 1

    if [[ -n $ERR ]]; then
      if [[ -n $RED ]]; then
        COLOR=$RED
      fi
    fi
    if [[ -n $WARN ]]; then
      if [[ -n $YEL ]]; then
        COLOR=$YEL
      fi
    fi
     [[ -z "$COLOR" ]] && NC=''

    if [[ -n "$PRINT_ONLY" ]]; then
      MESS=$(GetText "$MESS")
      if [[ -n $ERR ]]; then
        >&2 echo -e "${COLOR}$MESS${NC}"
      else
        echo -e "${COLOR}$MESS${NC}"
      fi
      return
    else
       echo -e "$MESS" >>"$LOG_FILE"
    fi

    if [[ -n "$PRINT" ]]; then
      MESS=$(GetText "$MESS")
      if [[ -n $ERR ]]; then
        >&2 echo -e "${COLOR}${MESS}${NC}"
      else
        echo -e "${COLOR}${MESS}${NC}"
      fi
    fi
} # LogMess()

# ------------------------------------------------------------------------------
SetColor()
{
  RED='\033[1;31m'
  YEL='\033[1;33m'
#  BLU='\033[1;34m'
#  GREEN='\033[0;32m'
  NC='\033[0m' # No Color
}

# ------------------------------------------------------------------------------
ContinueAsk()
{
  local PMT=" Continue execution (y/n)?" ANSW

  [[ "$AUTO_YES" = "Y" ]] &&  return 0

  while true; do
    read -rp "$PMT" ANSW
      case "$ANSW" in
        [Yy]* ) return 0  ;;
        [Nn]* )  JoinExit 1  ;;
        * ) echo " Answer 'yes' or 'no'"  ;;
      esac
  done
}

# ------------------------------------------------------------------------------
ReadPass()
{
  local PASS=''
  while true; do
    read -rsp "$(GetText  "Enter domain administrator password:")"  PASS
    if [[ -n "$PASS" ]]; then
      echo "$PASS"
      return
    fi
  done
}

# ------------------------------------------------------------------------------
RemoveSSSDCache()
{
  LogMess -D -m "RemoveSSSDCache() start"
  LogMess -p -m "Clear sssd cache"
  systemctl stop sssd || JoinExit 1
  rm -rf /var/lib/sss/{db,mc}/*
  systemctl start sssd || JoinExit 1
  rm -rf /tmp/krb5cc_*
  LogMess -D
}

# ------------------------------------------------------------------------------
DelPC()
{
  local ADMIN_NAME='' PC_NAME='' DOMAIN_NAME='' DC_NAME='' ADMIN_PASS='' J_COUNT

  while getopts "u:n:d:c:p:" OPTION; do
    case $OPTION in
      'u') ADMIN_NAME="$OPTARG" ;;
      'n') PC_NAME="$OPTARG" ;;
      'd') DOMAIN_NAME="$OPTARG" ;;
      'c') DC_NAME="-S $OPTARG" ;;
      'p') ADMIN_PASS="$OPTARG" ;;
      *) echo "DelPC() Invalid option: $*" ;;
    esac
  done
  shift $((OPTIND-1))
  OPTIND=1

  CheckPCInDomain -u "$ADMIN_NAME" -n "$PC_NAME" -d "$DOMAIN_NAME"  -p "$ADMIN_PASS" -N || return 0

  LogMess -d -m "DelPC() start"
  LogMess -p -m "Removing PC account from domain"

  LogMess -m "adcli delete-computer -v -U $ADMIN_NAME $DC_NAME --stdin-password --domain=$DOMAIN_NAME $PC_NAME"
  if adcli delete-computer -v -U "$ADMIN_NAME" $DC_NAME --stdin-password --domain="$DOMAIN_NAME" "$PC_NAME" <<< "$ADMIN_PASS" &>>"$LOG_FILE"; then
    J_COUNT=1
    while true; do
      sleep 5
      if ! CheckPCInDomain -u "$ADMIN_NAME" -n "$PC_NAME" -d "$DOMAIN_NAME"  -p "$ADMIN_PASS" -N ; then
        LogMess -d
        return 0
      fi
      J_COUNT=$((J_COUNT+1))

      if [ "$J_COUNT" -gt 15 ]; then
        LogMess -p -e -m 'Removing PC account from domain error, see $LOG_FILE'
        LogMess -d
        return 1
      fi
    done
  fi

  if cat "$LOG_FILE" | tail -n 5 | grep -q "No computer account for"; then
    LogMess -p -m 'Account $PC_NAME was not present in the domain!'
    return 0
  else
    if cat $LOG_FILE | tail -n 5 | grep -q "Preauthentication failed"; then
      LogMess -p -e -m "Invalid domain administrator name or password!"
    else
      LogMess -p -e -m 'Removing PC account from domain error, see $LOG_FILE'
    fi
  fi
  LogMess -d
  return 1
} # DelPC()

# ------------------------------------------------------------------------------
TymeSync()
{
  local DOMAIN_NAME="$1"
  LogMess -D -m "TymeSync() start"
  LogMess -p -m "Setting chronyd"
  LogMess -m "systemctl is-active --quiet systemd-timesyncd && systemctl disable systemd-timesyncd --now &>/dev/null"
  systemctl is-active --quiet systemd-timesyncd && systemctl disable systemd-timesyncd --now &>/dev/null

  sed -i '/server/d' $CHRONY_CONF_FILE
  sed -i '/pool/d' $CHRONY_CONF_FILE
  sed -i '/maxdistance/d' $CHRONY_CONF_FILE
  echo "pool $DOMAIN_NAME iburst prefer"  >>$CHRONY_CONF_FILE
  LogMess -m "systemctl restart chronyd"
  systemctl restart chronyd
  LogMess -D
}

# ------------------------------------------------------------------------------
GetDomainType()
{
  local DOMAIN_NAME="$1"
  local CHECK_OUT CHECK DOMAIN_TYPE='none'

  LogMess -d -m "GetDomainType() start"
  LogMess -m "realm discover $DOMAIN_NAME"

  CHECK_OUT=$(/bin/mktemp /tmp/domain_check.XXXXXX)
  DOMAIN_TYPE='ERR'

  if realm discover "$DOMAIN_NAME" &>"$CHECK_OUT"; then
    CHECK=$(cat "$CHECK_OUT")
    if grep -Eq "required-package:\s+.*ipa-client" <<< "$CHECK";  then
      DOMAIN_TYPE='IPA'
    else
      if grep -Eq "server-software:\s+ipa" <<< "$CHECK";  then
        DOMAIN_TYPE='IPA'
      else
        DOMAIN_TYPE='WIN'
      fi
    fi
  fi

  cat "$CHECK_OUT" &>>"$LOG_FILE"
  rm -f "$CHECK_OUT"
  echo "DOMAIN_TYPE: $DOMAIN_TYPE" &>>"$LOG_FILE"
  echo $DOMAIN_TYPE
  LogMess -d
  return 0
}

# ------------------------------------------------------------------------------
CheckDomain()
{
  local DOMAIN_NAME="$1"
  local DOMAIN_TYPE='' RET_CODE=1

  LogMess -d -m "CheckDomain() start"
  LogMess -m "adcli info $DOMAIN_NAME"
  if adcli info "$DOMAIN_NAME" &>>"$LOG_FILE"; then
    DOMAIN_TYPE=$(GetDomainType $DOMAIN_NAME)
    case $DOMAIN_TYPE in
      'IPA') RET_CODE=12 ;;
      'WIN') RET_CODE=11 ;;
      *) RET_CODE=13 ;;
    esac
  fi
  LogMess -d
  return $RET_CODE
}

# ------------------------------------------------------------------------------
CheckPCName()
{
  local PC_NAME="$1"

  if [[ $(grep -P '(^(?:[a-zA-Z0-9](?:(?:[a-zA-Z0-9\-]){0,14}[a-zA-Z0-9\-])+[a-zA-Z0-9])$)' <<< "$PC_NAME") && ${#PC_NAME} -le 15 ]]; then
    return 0
  fi
  LogMess -p -e -m 'Invalid PC name: $PC_NAME !'
  LogMess -P -m "Latin characters (A-Z,a-z), numbers (0-9) and hyphen (-) are allowed. The name should not exceed more than 15 characters."
  return 1
}

# ------------------------------------------------------------------------------
CheckPCInDomain()
{
  local ADMIN_NAME='' DOMAIN_NAME='' PC_NAME='' ADMIN_PASS=''  J_CHECK_OUT  CHECK IPA='' NO_CHK_DC=''

  while getopts "u:d:n:p:IN" OPTION; do
    case $OPTION in
      'u') ADMIN_NAME="$OPTARG" ;;
      'd') DOMAIN_NAME="$OPTARG" ;;
      'n') PC_NAME="$OPTARG" ;;
      'p') ADMIN_PASS="$OPTARG" ;;
      'I') IPA="Y" ;;
      'N') NO_CHK_DC="Y" ;;
      *) echo "CheckPCInDomain() Invalid option: $*" ;;
    esac
  done
  shift $((OPTIND-1))
  OPTIND=1

  LogMess -D -m "CheckPCInDomain() start"
  if [[ -z "$NO_CHK_DC" ]]; then
    GetPCInDC -d "$DOMAIN_NAME" -n "$PC_NAME" || return 12
  fi
  [[ -n "$IPA" ]] && return 199

  J_CHECK_OUT=$(/bin/mktemp /tmp/join_check.XXXXXX)

  LogMess -m "adcli show-computer -U $ADMIN_NAME --domain=$DOMAIN_NAME $PC_NAME --stdin-password"
  adcli show-computer -U "$ADMIN_NAME" --domain="$DOMAIN_NAME" "$PC_NAME" --stdin-password <<< "$ADMIN_PASS" &>"$J_CHECK_OUT"
  CHECK=$(cat "$J_CHECK_OUT")
  cat "$J_CHECK_OUT" &>>"$LOG_FILE"
  rm -f "$J_CHECK_OUT"

  if grep -Pq "sAMAccountName" <<< "$CHECK";  then
      LogMess -D
      return 0
  fi

  if grep -Pq "No computer account for" <<< "$CHECK";  then
      LogMess -D
      return 10
  fi

  if grep -Pq "Couldn't authenticate" <<< "$CHECK";  then
    LogMess -D
    return 11
  fi
  LogMess -D
  return 1
} # CheckPCInDomain()

# ------------------------------------------------------------------------------
PamWinbind()
{
  { sed -i -e 's\;cached_login\cached_login\g' $WINBIND_CONF_FILE;
  sed -i -e '/^cached_login/s/no/yes/g' $WINBIND_CONF_FILE;
  sed -i -e 's\;krb5_ccache_type =\krb5_ccache_type = FILE\g' $WINBIND_CONF_FILE;
  sed -i -e 's\;warn_pwd_expire\warn_pwd_expire\g' $WINBIND_CONF_FILE;
  sed -i -e 's\;krb5_auth\krb5_auth\g' $WINBIND_CONF_FILE;
  sed -i -e '/^krb5_auth/s/no/yes/g' $WINBIND_CONF_FILE; } &>>"$LOG_FILE"
}

# ------------------------------------------------------------------------------
GetDomainNameShort()
{
  local DOMAIN_NAME="$1"
  adcli info "$DOMAIN_NAME" | awk -F ' = ' '/domain-short/ {print $2}'
}

# ------------------------------------------------------------------------------
GetUpCase()
{
  local NAME="$1"
  echo "${NAME^^}"
}

# ------------------------------------------------------------------------------
GetLowCase()
{
  local NAME="$1"
  echo "${NAME,,}"
}

# ------------------------------------------------------------------------------
CreateKrb5()
{
  local DOMAIN_NAME='' DOMAIN_NAME_UP_CASE KRB5_DC1='' KRB5_DC2='' DC1=''

  while getopts "d:k:r:c:" OPTION; do
    case $OPTION in
      'd') DOMAIN_NAME="$OPTARG" ;;
      'k') KRB5_DC1="$OPTARG" ;;
      'r') KRB5_DC2="$OPTARG" ;;
      'c') DC1="$OPTARG" ;;
      *) echo "CreateKrb5() Invalid option: $*" ;;
    esac
  done
  shift $((OPTIND-1))
  OPTIND=1

  LogMess -D -m "CreateKrb5() start"
  DOMAIN_NAME_UP_CASE=$(GetUpCase "$DOMAIN_NAME")

  LogMess -m "Domain Information:"
  adcli info "$DOMAIN_NAME" &>>"$LOG_FILE"

  cat > $CRYPTO_POLICIES_FILE <<EOF
[libdefaults]
default_tgs_enctypes = aes256-cts-hmac-sha1-96 aes128-cts-hmac-sha1-96 RC4-HMAC DES-CBC-CRC DES3-CBC-SHA1 DES-CBC-MD5
default_tkt_enctypes = aes256-cts-hmac-sha1-96 aes128-cts-hmac-sha1-96 RC4-HMAC DES-CBC-CRC DES3-CBC-SHA1 DES-CBC-MD5
preferred_enctypes = aes256-cts-hmac-sha1-96 aes128-cts-hmac-sha1-96 RC4-HMAC DES-CBC-CRC DES3-CBC-SHA1 DES-CBC-MD5
EOF

cat > $KRB5_CONF_FILE <<EOF
includedir /etc/krb5.conf.d/

[logging]
    default = FILE:/var/log/krb5libs.log
    kdc = FILE:/var/log/krb5kdc.log
    admin_server = FILE:/var/log/kadmind.log

[libdefaults]
    default_realm = $DOMAIN_NAME_UP_CASE
    dns_lookup_realm = false
    dns_lookup_kdc = true
    ticket_lifetime = 24h
    renew_lifetime = 7d
    forwardable = true
    rdns = false
    pkinit_anchors = /etc/pki/tls/certs/ca-bundle.crt
    spake_preauth_groups = edwards25519
    default_ccache_name = FILE:/tmp/krb5cc_%{uid}
    default_tgs_enctypes = aes256-cts-hmac-sha1-96 aes128-cts-hmac-sha1-96 RC4-HMAC DES-CBC-CRC DES3-CBC-SHA1 DES-CBC-MD5
    default_tkt_enctypes = aes256-cts-hmac-sha1-96 aes128-cts-hmac-sha1-96 RC4-HMAC DES-CBC-CRC DES3-CBC-SHA1 DES-CBC-MD5
    preferred_enctypes = aes256-cts-hmac-sha1-96 aes128-cts-hmac-sha1-96 RC4-HMAC DES-CBC-CRC DES3-CBC-SHA1 DES-CBC-MD5

[realms]
$DOMAIN_NAME_UP_CASE = {
    $KRB5_DC1
    $KRB5_DC2
    admin_server = $DC1
    default_domain = $DOMAIN_NAME
}

[domain_realm]
.$DOMAIN_NAME = $DOMAIN_NAME_UP_CASE
$DOMAIN_NAME = $DOMAIN_NAME_UP_CASE
EOF
  LogMess -D
} # CreateKrb5()

# ------------------------------------------------------------------------------
JoinFreeIPA()
{
  local ADMIN_NAME='' PC_NAME='' DOMAIN_NAME='' ADMIN_PASS='' HOST_NAME DC IPA_REZ

  while getopts "u:n:d:p:" OPTION; do
    case $OPTION in
      'u') ADMIN_NAME="$OPTARG" ;;
      'n') PC_NAME="$OPTARG" ;;
      'd') DOMAIN_NAME="$OPTARG" ;;
      'p') ADMIN_PASS="$OPTARG" ;;
      *) echo "JoinFreeIPA() Invalid option: $*" ;;
    esac
  done
  shift $((OPTIND-1))
  OPTIND=1

  LogMess -D -m "JoinFreeIPA() start"

  DC=$(adcli info "$DOMAIN_NAME" | grep "domain-controller =" | awk '{print $3}')
  TymeSync "$DOMAIN_NAME"

  SetHostname ${PC_NAME}.$DOMAIN_NAME
  LogMess -p -m "Join domain"
  LogMess -m "ipa-client-install --mkhomedir --enable-dns-updates --domain=$DOMAIN_NAME --hostname $HOST_NAME --ntp-server=$DC -p $ADMIN_NAME -w passwd -U"
  ipa-client-install --mkhomedir --enable-dns-updates --domain="$DOMAIN_NAME" --hostname "$HOST_NAME" --ntp-server="$DC" -p "$ADMIN_NAME" -w "$ADMIN_PASS" -U &>>"$LOG_FILE"
  sed -i 's;default_ccache_name = KEYRING:persistent:%{uid};default_ccache_name = FILE:/tmp/krb5cc_%{uid};g' $KRB5_CONF_FILE

  IPA_REZ=$(tail -n1 /var/log/ipaclient-install.log | awk '{ print $NF }')
  if [[ "$IPA_REZ" = "successful" ]]; then
    LogMess -p -m "Computer has been successfully added to IPA domain"
    LogMess -D
    return 0
  else
    LogMess -p -e -m 'Error entering IPA domain, see /var/log/ipaclient-install.log'
    LogMess -D
    return 1
  fi
} # JoinFreeIPA()

# ------------------------------------------------------------------------------
DNSDelete()
{
  local DOMAIN DNS SHORT_NAME SHORT_NAME_UP_CASE HOST_NAME IP ADMIN_NAME='' ADMIN_PASS=''

  while getopts "u:p:" OPTION; do
    case $OPTION in
      'u') ADMIN_NAME="$OPTARG" ;;
      'p') ADMIN_PASS="$OPTARG" ;;
      *) echo "DNSDelete() Invalid option: $*" ;;
    esac
  done
  shift $((OPTIND-1))
  OPTIND=1

  LogMess -d -m "DNSDelete() start"
  DOMAIN=$(domainname -d)
  DNS=$(nmcli dev show | awk '/DNS/{print $2}' | head -n1)
  HOST_NAME=$(hostname -s)
  IP=$(hostname -I | awk '{print $1}')
  # IP=$(nslookup `hostname -s` | awk '/^Address: / {print $2}')

  LogMess -m "DNS: $DNS"
  LogMess -m "DOMAIN: $DOMAIN"
  LogMess -m "IP: $IP"
  LogMess -m "ADMIN_NAME: $ADMIN_NAME"

  SHORT_NAME=$(hostname | cut -d. -f2)
  SHORT_NAME_UP_CASE=$(GetUpCase "$SHORT_NAME")
  LogMess -m "samba-tool dns delete $DNS $DOMAIN $HOST_NAME A $IP -U $SHORT_NAME_UP_CASE\\${ADMIN_NAME}%password -d 3"
  if ! samba-tool dns delete "$DNS" "$DOMAIN" "$HOST_NAME" A "$IP" -U "$SHORT_NAME_UP_CASE\\${ADMIN_NAME}%$ADMIN_PASS" -d 3 &>>"$LOG_FILE"; then
    LogMess -m "samba-tool dns delete error"
    return 1
  fi
  LogMess -d
  return 0
} # DNSDelete()

# ------------------------------------------------------------------------------
ExitDomain()
{
  local ADMIN_NAME='' DC_NAME='' ADMIN_PASS='' DC_OPT='' DOMAIN HOST_NAME IPA OUT_IPA_1 OUT_IPA_2 LOCAL=''

  while getopts "u:c:fp:c:l" OPTION; do
    case $OPTION in
      'u') ADMIN_NAME="$OPTARG" ;;
      'c') DC_NAME="$OPTARG" ;;
      'l') LOCAL='y' ;;
      'p') ADMIN_PASS="$OPTARG" ;;
      *) echo "ExitDomain() Invalid option: $*" ;;
    esac
  done
  shift $((OPTIND-1))
  OPTIND=1

  LogMess -D -m "ExitDomain() start"
  IPA=$(realm list | grep server-software |  awk '{ print $NF }')

  DOMAIN=$(domainname -d)
  HOST_NAME=$(hostname -s)
  LogMess -p -m 'Computer $HOST_NAME in domain $DOMAIN'

  if [[ "$IPA" = "ipa" ]];  then
    ipa-client-install --uninstall -U &>> "$LOG_FILE"
    OUT_IPA_1=$(tail -n1 /var/log/ipaclient-uninstall.log | awk '{ print $NF }')
    OUT_IPA_2=$(grep 'Client uninstall complete' /var/log/ipaclient-uninstall.log | awk  '{ print $NF }')
    if [[ "$OUT_IPA_1" = "successful" || "$OUT_IPA_2" = "complete." ]]; then
      :
    else
      LogMess -p -e -m 'Error exiting IPA domain, see /var/log/ipaclient-uninstall.log';  LogMess -D
      return 1
    fi
  fi

  if [[ -z "$LOCAL" ]]; then
    [[ -n "$DC_NAME" ]] && DC_OPT="-c $DC_NAME"
    DelPC -d "$DOMAIN" -n "$HOST_NAME" -u "$ADMIN_NAME" -p "$ADMIN_PASS" $DC_OPT
  fi

  DNSDelete -u "$ADMIN_NAME" -p "$ADMIN_PASS"

  { realm leave -v --client-software=sssd; realm leave -v --client-software=winbind;  sss_cache -E;  kdestroy -A;  } &>>"$LOG_FILE"
  rm -rf /tmp/krb5cc* /var/lib/sss/db/*
  RestoreConfig
  LogMess -p -m 'Computer $HOST_NAME has been removed from domain.'
  LogMess -D
  return 0
} # ExitDomain()

# ------------------------------------------------------------------------------
SetAuthSelect()
{
  local MODE="$1"
  local AUTHSELECT_VER AUTHSELECT_VER_N

  [[ "$MODE" !=  "ver" ]] && LogMess -d -m "SetAuthSelect() start"
  AUTHSELECT_VER=$(rpm -q --queryformat '%{VERSION}' authselect)
  AUTHSELECT_VER_N=${AUTHSELECT_VER//./}

  if [[ "$AUTHSELECT_VER_N" -ge 150 ]]; then
    [[ "$MODE" =  "ver" ]] && return 0
    LogMess -m "New version authselect $AUTHSELECT_VER"
    if [[ "$MODE" =  "winbind" ]]; then
      [[ -d "/etc/authselect/custom/winbind_domain" ]] && rm -rf "/etc/authselect/custom/winbind_domain"
      { authselect create-profile winbind_domain -b winbind;
      sed -i 's/\bhosts:.*/hosts:      files myhostname resolve dns mdns4_minimal [!UNAVAIL=return]/g' $WINBIND_NSSWITCH_CONF_FILE;
      authselect select custom/winbind_domain with-krb5 with-faillock with-fingerprint with-mkhomedir --force; } &>>"$LOG_FILE"
    else
      [[ -d "/etc/authselect/custom/sssd_domain" ]] && rm -rf "/etc/authselect/custom/sssd_domain"
      { authselect create-profile sssd_domain -b sssd;
      sed -i 's/\bhosts:.*/hosts:      files myhostname resolve dns mdns4_minimal [!UNAVAIL=return]/g' $SSSD_NSSWITCH_CONF_FILE;
      authselect select custom/sssd_domain with-faillock with-fingerprint with-smartcard with-mkhomedir --force; } &>>"$LOG_FILE"
    fi
  else
    [[ "$MODE" =  "ver" ]] && return 1
    if [[ "$MODE" =  "winbind" ]]; then
      { authselect select winbind with-mkhomedir with-krb5 --force; authselect enable-feature with-mkhomedir; } &>>"$LOG_FILE"
    else
      authselect select sssd with-fingerprint with-gssapi with-mkhomedir with-smartcard --force &> /dev/null
      sed -i 's/\bhosts:.*/hosts:      files dns resolve [!UNAVAIL=return] myhostname mdns4_minimal/g' $USER_NSSWITCH_CONF_FILE &>>"$LOG_FILE"
      authselect apply-changes &>>"$LOG_FILE"
    fi
  fi
  LogMess -d
} # SetAuthSelect()

# ------------------------------------------------------------------------------
JoinWinbind()
{
  local DOMAIN_NAME='' ADMIN_NAME='' ADMIN_PASS='' OS_NAME='' OS_VERSION='' OU_JOIN='' WG='' DC1='' SHORT_DOMEN DOMAIN_NAME_UP_CASE

  while getopts "d:u:p:o:v:j:w:c:" OPTION; do
    case $OPTION in
      'd') DOMAIN_NAME="$OPTARG" ;;
      'u') ADMIN_NAME="$OPTARG" ;;
      'p') ADMIN_PASS="$OPTARG" ;;
      'o') OS_NAME="$OPTARG" ;;
      'v') OS_VERSION="$OPTARG" ;;
      'j') OU_JOIN="$OPTARG" ;;
      'w') WG="-W $OPTARG" ;;
      'c') DC1="$OPTARG" ;;
      *) echo "JoinWinbind() Invalid option: $*" ;;
    esac
  done
  shift $((OPTIND-1))
  OPTIND=1

  LogMess -D -m "JoinWinbind() start"
  LogMess -p -m "Join domain"
  sed -i "s/SELINUX=enforcing/SELINUX=permissive/" $SELINUX_CONFIG_FILE
  setenforce 0 &>>"$LOG_FILE"
  systemctl disable sssd &>>"$LOG_FILE"
  systemctl stop sssd &>>"$LOG_FILE"

  LogMess -p -m "Join domain (winbind)"
  LogMess -m "realm join -vvv -U $ADMIN_NAME --client-software=winbind $DOMAIN_NAME --os-name=$OS_NAME --os-version=$OS_VERSION $OU_JOIN"
  if ! realm join -vvv -U "$ADMIN_NAME" --client-software=winbind "$DOMAIN_NAME" --os-name="$OS_NAME" --os-version="$OS_VERSION" $OU_JOIN  <<< "$ADMIN_PASS" &>>"$LOG_FILE"; then
    LogMess -p -e -m 'Join domain error, see $LOG_FILE'
    LogMess -D
    return 1
  fi

  LogMess -p -m "Run authselect"
  SetAuthSelect 'winbind'

  mkdir -p /var/lib/domain/
  mkdir -p /var/lib/domain/run

  SHORT_DOMEN=$(GetDomainNameShort "$DOMAIN_NAME")
  DOMAIN_NAME_UP_CASE=$(GetUpCase "$DOMAIN_NAME")

  LogMess -p -m "Setting samba"
  cat >$SMB_CONF_FILE <<EOF
[global]
    workgroup = $SHORT_DOMEN
    realm = $DOMAIN_NAME_UP_CASE
    security = ADS

    winbind enum groups = Yes
    winbind enum users = Yes
    winbind offline logon = Yes
    winbind use default domain = no

    winbind refresh tickets = Yes
    winbind cache time = 300
    wins support = no

    idmap cache time = 900
    idmap config $SHORT_DOMEN : backend = sss
    idmap config * : range = 200000-2147483647

    client min protocol = NT1
    client max protocol = SMB3
    kerberos method = secrets and keytab
    template homedir = /home/%U@%D

    template shell = /bin/bash
    nt pipe support = no
    machine password timeout = 0
    vfs objects = acl_xattr
    map acl inherit = yes
    store dos attributes = yes

    printing = cups
    printcap name = cups
    load printers = yes
    cups options = raw

[homes]
    comment = Home Directories
    valid users = %S, %D%w%S
    browseable = No
    read only = No
    inherit acls = Yes

[printers]
    comment = All Printers
    path = /var/tmp
    printable = Yes
    create mask = 0600
    browseable = No

[print$]
    comment = Printer Drivers
    path = /var/lib/samba/drivers
    write list = @printadmin root
    force group = @printadmin
    create mask = 0664
    directory mask = 0775'
EOF

  LogMess -p -m "Samba configuration test"
  echo -e "\n" | testparm &>>"$LOG_FILE"

  cat >$LIMITS_FILE <<EOF
*     -  nofile  16384
root  -  nofile  16384
EOF

  LogMess -p -m "Join  samba domain and registering DNS record"
  LogMess -m "net ads join -S $DC1 -U ${ADMIN_NAME}%password $WG"
  if ! net ads join -S "$DC1" -U "${ADMIN_NAME}%${ADMIN_PASS}" $WG &>>"$LOG_FILE"; then
    LogMess -p -e -m 'Run net ads join error, see $LOG_FILE'
    return 1
  fi

  LogMess -p -m "Starting winbind service"
  if ! systemctl enable winbind --now &>>"$LOG_FILE"; then
    LogMess -p -e -m 'Starting winbind service error, see $LOG_FILE'
    return 1
  fi

  LogMess -p -m "Starting smb service"
  if ! systemctl enable smb --now &>>"$LOG_FILE"; then
    LogMess -p -e -m 'Starting smb service error, see $LOG_FILE'
    return 1
  fi
  PamWinbind
  LogMess -p -m "Computer has been successfully joined to domain"
  LogMess -D
  return 0
} # JoinWinbind()

# ------------------------------------------------------------------------------
Join()
{
  local DOMAIN_NAME='' ADMIN_NAME='' ADMIN_PASS='' OS_NAME='' OS_VERSION='' OU_JOIN='' J_COUNT SRV_DC RET_CODE SAVE_CASE='' LOWER_CASE='' LOGIN_CASE DNS_AUTH_NONE SSH_KRB5 DC1='' DC2='' FQN='False'  SHORT_DOMEN DOMAIN_NAME_UP_CASE DOMAIN DNS_AUTH=''

  while getopts "d:u:p:o:v:j:siakc:e:q" OPTION; do
    case $OPTION in
      'd') DOMAIN_NAME="$OPTARG" ;;
      'u') ADMIN_NAME="$OPTARG" ;;
      'p') ADMIN_PASS="$OPTARG" ;;
      'o') OS_NAME="$OPTARG" ;;
      'v') OS_VERSION="$OPTARG" ;;
      'j') OU_JOIN="$OPTARG" ;;
      's') SAVE_CASE='Y' ;;
      'i') LOWER_CASE='Y' ;;
      'a') DNS_AUTH_NONE='Y' ;;
      'k') SSH_KRB5='Y' ;;
      'c') DC1="$OPTARG" ;;
      'e') DC2="$OPTARG" ;;
      'q') FQN='True' ;;
      *) echo "Join() Invalid option: $*" ;;
    esac
  done
  shift $((OPTIND-1))
  OPTIND=1

  LogMess -D -m "Join() start"
  SRV_DC="$DC1"
  DNSDelete -u "$ADMIN_NAME" -p "$ADMIN_PASS"

  J_COUNT=1
  LogMess -p -m "Join domain"

  LogMess -m "join -vvv -U $ADMIN_NAME $SRV_DC $OU_JOIN --os-name=$OS_NAME --os-version=$OS_VERSION"
  realm join -vvv -U "$ADMIN_NAME" "$SRV_DC" $OU_JOIN --os-name="$OS_NAME" --os-version="$OS_VERSION" <<< "$ADMIN_PASS" &>>"$LOG_FILE"
  RET_CODE=$?
  while [[ $RET_CODE -ne 0 ]]; do
    J_COUNT=$((J_COUNT+1))
    GetText "Join domain,  attempt"; echo " $J_COUNT"
    LogMess -m "Join domain,  attempt $J_COUNT"
    LogMess -m "join -vvv -U $ADMIN_NAME $SRV_DC $OU_JOIN --os-name=$OS_NAME --os-version=$OS_VERSION"
    realm join -vvv -U "$ADMIN_NAME" "$SRV_DC" $OU_JOIN --os-name="$OS_NAME" --os-version="$OS_VERSION" <<< "$ADMIN_PASS" &>>"$LOG_FILE"
    RET_CODE=$?
    [ $RET_CODE -ne 0 ] && SRV_DC="$DOMAIN_NAME"
    sleep 1
    if [ "$J_COUNT" -gt 3 ]; then
      LogMess -p -e -m 'Join domain error, see $LOG_FILE'
      LogMess -D
      return 1
    fi
  done

  if [[ -n "$SAVE_CASE" ]]; then
    LOGIN_CASE="Preserving"
  elif [[ -n "$LOWER_CASE" ]]; then
    LOGIN_CASE="False"
  elif [[ -z "$SAVE_CASE" || -z "$LOWER_CASE" ]]; then
    LOGIN_CASE="False"
  fi

  if [[ -n "$DNS_AUTH_NONE" ]]; then
    DNS_AUTH="none"
  else
    DNS_AUTH="GSS-TSIG"
  fi

  SHORT_DOMEN=$(GetDomainNameShort "$DOMAIN_NAME")
  DOMAIN_NAME_UP_CASE=$(GetUpCase "$DOMAIN_NAME")
  DOMAIN=$(domainname -d)

  SetAuthSelect "ver" && authselect select custom/sssd_domain with-faillock with-fingerprint with-smartcard with-mkhomedir --force &>>"$LOG_FILE"

  LogMess -p -m "Setting sssd"
  cat >$SSSD_CONF_FILE <<EOF
[sssd]
domains = $DOMAIN
config_file_version = 2
services = nss, pam

# re_expression = (?P<name>[a-z0-9._-]+)
# re_expression = (?P<name>[a-z0-9._-]+(?![@]))

[domain/${DOMAIN}]
ad_domain = $DOMAIN
ad_server = '$DC1''$DC2'
krb5_realm = $DOMAIN_NAME_UP_CASE
case_sensitive = $LOGIN_CASE
realmd_tags = manages-system joined-with-samba

#krb5_use_subdomain_realm=True
cache_credentials = True
ad_update_samba_machine_account_password = true

id_provider = ad
access_provider = ad
krb5_store_password_if_offline = True
default_shell = /bin/bash
ldap_id_mapping = True
ad_gpo_access_control = disabled
dns_resolver_timeout = 10

use_fully_qualified_names = $FQN
fallback_homedir = /home/%u@%d

#access_provider = simple
#simple_allow_users = user1@example.com, user2@example.com
#simple_allow_groups = group@example.com

enumerate = false
# ignore_group_members = True
ldap_referrals = false

dyndns_update = true
dyndns_refresh_interval = 43200
dyndns_update_ptr = true
dyndns_ttl = 3600
dyndns_auth = $DNS_AUTH

krb5_lifetime = 24h
krb5_renewable_lifetime = 7d
krb5_renew_interval = 300s

[nss]
#entry_cache_timeout = 15
#get_domains_timeout = 10
EOF

  LogMess -p -m "Setting limits"
  cat >$LIMITS_FILE <<EOF
*     -  nofile  16384
root  -  nofile  16384
EOF

  LogMess -p -m "Setting samba"
  cat >$SMB_CONF_FILE <<EOF
[global]
    workgroup = $SHORT_DOMEN
    realm = $DOMAIN_NAME_UP_CASE
    security = ADS
    passdb backend = tdbsam

    winbind enum groups = Yes
    winbind enum users = Yes
    winbind offline logon = Yes
    winbind use default domain = No
    winbind refresh tickets = Yes

    idmap cache time = 900
    idmap config * : backend = tdb
    idmap config * : range = 10000-99999
    idmap config $SHORT_DOMEN : backend = rid
    idmap config $SHORT_DOMEN : range = 100000-999999

    client min protocol = NT1
    client max protocol = SMB3

    dedicated keytab file = /etc/krb5.keytab
    kerberos method = secrets and keytab

    machine password timeout = 60
    vfs objects = acl_xattr
    map acl inherit = yes
    store dos attributes = yes

    printing = cups
    printcap name = cups
    load printers = yes
    cups options = raw

[homes]
    comment = Home Directories
    valid users = %S, %D%w%S
    browseable = No
    read only = No
    inherit acls = Yes

[printers]
    comment = All Printers
    path = /var/tmp
    printable = Yes
    create mask = 0600
    browseable = No

[print$]
    comment = Printer Drivers
    path = /var/lib/samba/drivers
    write list = @printadmin root
    force group = @printadmin
    create mask = 0664
    directory mask = 0775
EOF

  PamWinbind

  LogMess -p -m "Starting sssd service"
  systemctl enable sssd &>>"$LOG_FILE"
  systemctl stop sssd &>>"$LOG_FILE"
  if ! systemctl start sssd &>>"$LOG_FILE"; then
    LogMess -p -e -m 'Starting sssd service error, see $LOG_FILE'
    return 1
  fi

  LogMess -p -m "Join  samba domain and registering DNS record"
  LogMess -m "net ads join -S $DC1 -U ${ADMIN_NAME}%password -D $DOMAIN_NAME -d 2"
  if ! net ads join -S "$DC1" -U "${ADMIN_NAME}%$ADMIN_PASS" -D "$DOMAIN_NAME" -d 2 &>>"$LOG_FILE"; then
    LogMess -p -e -m 'Run net ads join error, see $LOG_FILE'
    if tail -n 4 "$LOG_FILE" | grep -q "Failed to join domain"; then
      LogMess -p -e -m 'Computer is joined to the domain, but an error occurred while executing  command, see $LOG_FILE'
    fi
    LogMess -D
    return 1
  fi

  if [[ -n "$SSH_KRB5" ]]; then
    sed -i 's/^#KerberosAuthentication .*/KerberosAuthentication yes/' $SSHD_CONFIG_FILE
    sed -i 's/^#KerberosTicketCleanup .*/KerberosTicketCleanup yes/' $SSHD_CONFIG_FILE
    sed -i 's/^#GSSAPIAuthentication .*/GSSAPIAuthentication yes/' $SSHD_CONFIG_FILE
    sed -i 's/^#GSSAPICleanupCredentials .*/GSSAPICleanupCredentials yes/' $SSHD_CONFIG_FILE
    sed -i 's/^#.*GSSAPIAuthentication .*/GSSAPIAuthentication yes/' $SSH_CONFIG_FILE
    sed -i 's/^#.*GSSAPIDelegateCredentials .*/GSSAPIDelegateCredentials yes/' $SSH_CONFIG_FILE
  fi

  LogMess -p -m "Computer has been successfully joined to domain"
  LogMess -D
  return 0
} # Join()

# ------------------------------------------------------------------------------
GetPCInDC()
{
  local DOMAIN='' PC_NAME=''

  while getopts "d:n:" OPTION; do
    case $OPTION in
      'd') DOMAIN="$OPTARG" ;;
      'n') PC_NAME="$OPTARG" ;;
      *) echo "GetPCInDC() Invalid option: $*" ;;
    esac
  done
  shift $((OPTIND-1))
  OPTIND=1

  LogMess -p -m "Check that computer name does not match the domain controller name"

  for FDC in $(adcli info "$DOMAIN" | grep "domain-controllers =" | sed s'/domain-controllers =//g' | tee -a "$LOG_FILE"); do
    SHORT_DC=$(cut -d'.' -f1 <<< "$FDC")
    if grep -Eqi "^${SHORT_DC}$" <<<$PC_NAME; then
      LogMess -p -e -m "The PC name must not match the domain controller name!"
      return 1
    fi
  done
  return 0
}

# ------------------------------------------------------------------------------
GetDCNames()
{
  local DOMAIN='' FL=''
  DC1='' DC2='' KRB5_DC1='' KRB5_DC2=''

  while getopts "d:" OPTION; do
    case $OPTION in
      'd') DOMAIN="$OPTARG" ;;
      *) echo "GetDCNames() Invalid option: $*" ;;
    esac
  done
  shift $((OPTIND-1))
  OPTIND=1

  LogMess -p -m "Get domain controllers names"

  for FDC in $(adcli info "$DOMAIN" | grep "domain-controllers =" | sed s'/domain-controllers =//g' | tee -a "$LOG_FILE"); do
    FL=$(adcli info --domain-controller="$FDC" 2>/dev/null | grep "domain-controller-flags =" 2>/dev/null | tee -a "$LOG_FILE")
    if  grep -Eq "\s+pdc\s+" <<<"$FL"; then
      KRB5_DC1="kdc = $FDC"
      DC1="$FDC"
    else
      if  grep -Eq "\s+closest\s+" <<<"$FL"; then
        KRB5_DC2="kdc = $FDC"
        DC2=", $FDC, _srv_"
      fi
    fi
  done

  if [[ -z "$DC1" ]]; then
    LogMess -p -e -m 'Get domain controllers names error, see $LOG_FILE'
  fi
  return 0
}

# ------------------------------------------------------------------------------
SetHostname()
{
  local OLD_NAME='' NEW_NAME=''
  local NAME="$1"

  OLD_NAME=$(hostname)

  [[ "$OLD_NAME" =  "$NAME" ]] && return 0

  LogMess -p -m "Set hostname"
  echo -e "127.0.0.1 $NAME" >>$HOSTS_FILE

  hostnamectl set-hostname "$NAME"

  NEW_NAME=$(hostname); export NEW_NAME
  LogMess -p -m 'New PC name: $NEW_NAME'
  sed -i -E "s/^.+${OLD_NAME}.*$//g" $HOSTS_FILE
}

# ------------------------------------------------------------------------------
SaveConfig()
{
  LogMess -d -m "SaveConfig:"
  for FILE in $CH_FILES; do
    SAVE_FILE="${FILE}_$SAVE_SUFFIX"
    if [[ -f "$FILE" ]]; then
      if [[ ! -f "$SAVE_FILE" ]]; then
        LogMess -m "cp -a $FILE $SAVE_FILE"
        cp -a "$FILE" "$SAVE_FILE"
      fi
    fi
  done
  LogMess -d
}

# ------------------------------------------------------------------------------
RestoreConfig()
{
  local N_HOSTNAME
  LogMess -d -m "RestoreConfig:"
  for FILE in $CH_FILES; do
    SAVE_FILE="${FILE}_$SAVE_SUFFIX"
    if [[ -f "$SAVE_FILE" ]]; then
      LogMess -m "mv -f $SAVE_FILE $FILE"
      mv -f "$SAVE_FILE" "$FILE"
    fi
  done

  N_HOSTNAME=$(cat /etc/hostname  | xargs)
  hostnamectl set-hostname "$N_HOSTNAME"

  LogMess -m "systemctl restart chronyd"
  systemctl restart chronyd
  LogMess -d
}

# ------------------------------------------------------------------------------
JoinExit()
{
  local ST="$1"
  if [[ $ST -ne 0 ]]; then
    RestoreConfig
    LogMess -m "Script completed with errors."
    LogMess -S
    exit $ST
  fi
  LogMess -m "Script completed successfully."
  LogMess -S
  exit $ST
}
